<table {% if table_style %}class="{{table_style}}"{% endif %}>
    <thead>
        <tr>
            {% for col in table.header %}<th>{{col}}</th>{% endfor %}
        </tr>
    </thead>
    <tbody>
    {% for row in table.rows %}<tr>
        <th>
        {% if not create_links or row.is_base %}
        {{row.header.label}}
        {% else %}
        <a href="?cut={{cell.drilldown(dimension, row.header.key)}}">
            {{row.header.label}}
        </a>
        {% endif %}
        </th>
        {% for tcell in row.data %}<td>{{tcell}}</td>{% endfor %}
    </tr>
    {% endfor %}
    </tbody>
</table>



